//******************************** CGI NovaCmd Info: ********************************//
//Commands Info:    queueCmds, sendCmds
//  queueCmds does not garuantee that the command will be sent and may return false 
//    if sending is busy, use this if you do not care if the data makes it through.
//    Hint: This is good for regularly scheduled cmds
//  sendCmds will ensure that the command gets sent to the unit and will continue 
//    to try to send it up failure, however calling this too many times too quickly 
//    can cause a queueing effect where the commands do not get sent as quickly as
//    they are scheduled, which causes problems.
//    Hint: This is good for event driven cmds
//
//Args Info:        cmd, timeout, blocking
//  cmd: this could be a single string or an array of strings (10 max)
//    if a single string is specified, then only one command will be sent
//    if a list is specified, then all the messages in the list will be sent with
//    the same request. If more than 10 are in the list, then queueCmds will 
//    only send the first 10. sendCmds will schedule the messages 10 at a time.
//  timeout: specifies how long the receiver should wait for the response to come
//    if not specified, the receiver will use its default, current set to 1.2 seconds
//  blocking: if false, then the receiver will not wait the entire timeout period
//    but will return the first response it receives.
//    If true, then the receiver will wait the entire timeout period regardless of
//    how many messages it receives.
//    Hint: use false with a small timeout for commands with ignored responses, or
//      with no responses, or with quick responses. Use false with a long timeout if
//      the response of the command takes a while to return.
//    Hint: use true with commands that will return large data with a pause inbetween
//      the data transmission, or when sending multiple commands that will have their
//      own response.
//    Hint: Try to avoid using large timeout with blocking option as much as possible
//      because such combination will hold up the queue on the receiver end
//**********************************************************************************//

//Globals
var requesters = [];
var sequenceRequester;
var page;
var msgs = [];
var waitingTime = 2000;
var maxCmds = 10;
var maxRequesters = 20;
var MAX_BIN_LEN = 2048;

//*********************************************************************************************************
//************************************ Request Handler Class **********************************************
//*********************************************************************************************************

function initRequestListeners(pageIn, msgIn)
{
   //start by creating 4 requesters
   for(var i=0; i<4; i++)
      requesters.push(new RequestHandler());
      
   sequenceRequester = new RequestHandler();
   
   page = pageIn;
   if(msgIn)
      addListenerObj(msgIn);
      
   runSenders();
}

function addListenerObj(msgIn)
{
   msgs.push(msgIn);
}

function getAvailableRequester()
{
   for(var i=0; i<requesters.length; i++)
   {
      if(requesters[i].isReady())
      {
         return requesters[i];
      }
   }
   
   return null;
}

function RequestHandler()
{
   var handler = null;
   if (typeof XMLHttpRequest != 'undefined')
      handler =  new XMLHttpRequest();
   else if (window.ActiveXObject) {
      var avers = ["Microsoft.XmlHttp", "MSXML2.XmlHttp", "MSXML2.XmlHttp.3.0", "MSXML2.XmlHttp.4.0", "MSXML2.XmlHttp.5.0"];
      for (var i = avers.length - 1; i >= 0; i--) {
         try {
            handler = new ActiveXObject(avers[i]);
         } catch (e) { }
      }
   }
   
   var updateForm = function()
   {
      if (this.readyState==4 && this.status==200)
      {
         var i;
         for(i = 0; i < msgs.length; i++)
         {
            if( msgs[i].update(this.responseText) )
            {
               page.format(msgs[i]);
            }
         }
         gotResponse = true;
      }
   };
   
   if(handler)
   {
      var lastSentRequest = (new Date()).getTime()/1;
      var gotResponse = true;
      handler.onreadystatechange = updateForm;
   }
   else
   {
      throw new Error('XMLHttp (AJAX) not supported');
   }
   
   this.getHandler = function()
   {
      return handler;
   };

   this.isReady = function()
   {
      return (gotResponse || ((new Date()).getTime() - lastSentRequest) > waitingTime);
   };

   var getCmdStr = function(cmd)
   {
      //change to ascii hex representation
      cmd = cmd.replace(/\s/g, "%20");
      cmd = cmd.replace(/\?/g,  "%3F");
      return "?novacmdname=" + cmd;
   };

   this.sendCgiRequest = function(request)
   {
      // true is asynchronous connection
      handler.open("GET", request, true);
      handler.onreadystatechange = updateForm;
      handler.send();
      
      lastSentRequest = (new Date()).getTime()/1;
      gotResponse = false;
      
      return true;
   };

   this.sendNovaCommand = function(cmd, timeOut, blocking)
   {
      var timeOutStr = "";
      var blockingStr = "";
      var cmdStr = "";
      if(timeOut)
         timeOutStr = "&timeout=" + timeOut;
      if(blocking)
         blockingStr = "&blocking=" + ((blocking) ? '1' : '0');
         
      if(typeof cmd == 'string')
      {
         cmdStr += getCmdStr(cmd);
      }
      else
      {
         for(var cmdi=0; cmdi<cmd.length && cmdi<maxCmds; cmdi++)
         {
            cmdStr += getCmdStr(cmd[cmdi]);
         }
      }
      
      return this.sendCgiRequest("/novacmd.cgi" + cmdStr + timeOutStr + blockingStr);
   };
   
   this.sendBinary = function(url, cmdArray, offset, cmdLen, holdForResp)
   {
      var ajaxSend = (holdForResp) ? false : true;
      ajaxSend = true; // Can only be true
      handler.open("POST", url, ajaxSend);
      //handler.setRequestHeader("Content-Length", cmdLen);
      handler.onreadystatechange = updateForm;
      // cmdArray is a Uint8Array typed array but its length is
      // fixed to the value used in the constructor.  Even though 
      // Content-Length is set to the value of cmdLen the send method
      // still sends the full size of the cmdArray buffer and sends that
      // same value in Content-Length.  As a workaround if cmdLen does
      // not equal cmdArray.length create a new Uint8Array of the correct
      // size and copy the data to it.
      if (cmdLen != cmdArray.length || offset != 0) {
         var newArray = new Uint8Array(cmdLen);
         for (var i = 0; i < cmdLen; i++) {
            newArray[i] = cmdArray[i+offset];
         }
         handler.send(newArray.buffer);
      } else {
         handler.send(cmdArray.buffer);
      }
      
      lastSentRequest = (new Date()).getTime()/1;
      gotResponse = false;
   };
}

//*********************************************************************************************************
//************************************** Command Functions ************************************************
//*********************************************************************************************************
var commandBurstQueue = [];
var commandSequenceQueue = [];

var MessageType = { ASCII: 0, BINARY: 1 };

function runSenders()
{
   setInterval(
      function () {
         //burst dequeue - tries to send all the commands in the queue rightaway
         for(var count = 0; count < commandBurstQueue.length; count++)
         {
            var requester = getAvailableRequester();
            if(requester)
            {
               var cmdArgsBurst = commandBurstQueue.shift();
               if(cmdArgsBurst[0] == MessageType.ASCII) //is ascii
                  requester.sendNovaCommand(cmdArgsBurst[1], cmdArgsBurst[2], cmdArgsBurst[3]);
               else //is binary
                  requester.sendBinary(cmdArgsBurst[1], cmdArgsBurst[2], cmdArgsBurst[3], cmdArgsBurst[4], false);
            }
            //if we got here, that means we could use another requester
            else if(requesters.length < maxRequesters)
               requesters.push(new RequestHandler());
            else
               break;
         }
         
         //sequence dequeue - tries to send all the commands in the queue one at a time
         if(commandSequenceQueue.length > 0 && sequenceRequester.isReady())
         {
            var cmdArgsSequence = commandSequenceQueue.shift();
            if(cmdArgsSequence[0] == MessageType.ASCII) //is ascii
               sequenceRequester.sendNovaCommand(cmdArgsSequence[1], cmdArgsSequence[2], cmdArgsSequence[3]);
            else //is binary
               sequenceRequester.sendBinary(cmdArgsSequence[1], cmdArgsSequence[2], cmdArgsSequence[3], cmdArgsSequence[4], false);
         }
      },
      100
   );
}

//Sends all given commands at once with multiple CGI requests, not dependant on each other
function sendCmds(cmd, timeOut, blocking)
{
   var cmdList = (typeof cmd == 'string') ? [cmd] : cmd;
   for(var i=0; i<cmdList.length; i++)
      commandBurstQueue.push([MessageType.ASCII, cmdList[i], timeOut, blocking]);
}

//Sends all given commands at once with a single CGI requests, not dependant on each other
function sendGroup(cmd, timeOut, blocking)
{
   commandBurstQueue.push([MessageType.ASCII, cmd, timeOut, blocking]);
}

//Sends all the commands in order after adding them to a single queue, will send in the order received
function queueCmds(cmd, timeOut, blocking)
{
   var cmdList = (typeof cmd == 'string') ? [cmd] : cmd;
   for(var i=0; i<cmdList.length; i++)
      commandSequenceQueue.push([MessageType.ASCII, cmdList[i], timeOut, blocking]);
}

//Sends all the commands in order after adding them to a single queue, will send in the order received.
//Won't queue duplicate commands.  Checks before queuing and drops the command if it's already queued. 
function queueCmd(cmd, timeOut, blocking)
{
   var cmdList = (typeof cmd == 'string') ? [cmd] : cmd;

   for(var i=0; i<commandSequenceQueue.length; i++)
   {
      var queuedCmd = commandSequenceQueue[i];
      var newCmd    = cmdList[0];
      if (queuedCmd[1] === newCmd)
         return;
   }

   for(var i=0; i<cmdList.length; i++)
      commandSequenceQueue.push([MessageType.ASCII, cmdList[i], timeOut, blocking]);
}

//TODO: Handle failed cases
function sendBinary(url, cmdArray, cmdLen, holdForResp)
{
   //if we want to wait on a definite response (dangerous) then get a requester and send it directly
   if(holdForResp)
   {
      var requester = getAvailableRequester();
      if(requester)
      {
         requester.sendBinary(url, cmdArray, 0, cmdLen, holdForResp);
      }
   }
   //otherwise, let the sender task handle it
   else
   {
   //   for(var offset = 0; offset < cmdLen; offset += MAX_BIN_LEN)
   //      commandBurstQueue.push([MessageType.BINARY, url, cmdArray, offset, Math.min(cmdLen-offset, MAX_BIN_LEN)]);
        commandBurstQueue.push([MessageType.BINARY, url, cmdArray, 0, cmdLen]);
   }
}

function queueBinary(url, cmdArray, cmdLen, holdForResp)
{
   for(var offset = 0; offset < cmdLen; offset += MAX_BIN_LEN)
      commandSequenceQueue.push([MessageType.BINARY, url, cmdArray, offset, Math.min(cmdLen-offset, MAX_BIN_LEN)]);
}

//*********************************************************************************************************
//*************************************** Helper Functions ************************************************
//*********************************************************************************************************

//function to send false information to logout the current user
function logout()
{
   // true is asynchronous connection
   requesters[0].getHandler().open("GET", "/", true, "user", 'whatyoucallabadpassword');
   requesters[0].getHandler().send();
   document.execCommand('ClearAuthenticationCache', 'false');
}

//function to queue the current username logged in
function whoami()
{
   var requester = getAvailableRequester();
   if(requester)
   {
      requester.sendCgiRequest("/novacmd.cgi?syscmdname=whoami");
   }
}

//a short function to call document.getElementById()
function getById(str)
{
   return document.getElementById(str);
}

//*********************************************************************************************************
//*********************************** Standard Message Object *********************************************
//*********************************************************************************************************

//put the standard ascii parser with xmlHandler since every page could potentially use it
// this.input   = the entire message receieved
// this.name    = the name of the command, ie. [OUTPUT]
// this.msg     = the (first) message following the command name until \r\n, ie. PVT1B,ONCHANGE
// this.arg     = an array of the command args, ie. [PVT1B, ONCHANGE]
// this.msgList = an array of all the messages, incase there are more than one msg returned
// this.argList = an 2D array of args in each of the returned commands, ie. [[PVT1B, ONCHANGE],[MEAS1B, ONCHANGE]]
function standardMsg_obj(commandName)
{
   var MsgName = commandName;
   this.msg = "";
   
   this.update = function(input)
   {      
      var start = input.indexOf(MsgName);

      //don't start parsing if this is not the right message
      if(start < 0)
      {
         return false;
      }
      this.name = MsgName;
      this.input = input.substring(start);
      start += MsgName.length;
      
      //whatever you get, spit it back
      
      var msgLine = input.substring(start).split("\r\n");
      this.msg = msgLine[0]; //set the message up to the crlf
      this.arg = cleanUpArgList(this.msg.split(','));

      this.msgList = [];
      this.argList = [];
      for(var ii=0; ii<msgLine.length; ii++)
      {
         msgLine[ii] = msgLine[ii].replace(MsgName, "");
         this.msgList.push(msgLine[ii]);
         this.argList.push(cleanUpArgList(msgLine[ii].split(',')));
      }
      return true;
   };
   
   //firsts trims the leading/trailing spaces, and then gets rid of any quotes
   var cleanUpArgList = function(args)
   {
      for(var argi=0; argi<args.length; argi++)
         args[argi] = args[argi].replace(/^\s+|\s+$/g,'').replace(/"/g,'');
      return args;
   };
}
